(defclass human ()
  ((population :allocation :class :initform 0)
   (name :initarg :name :reader name :writer rename)
   (size :initarg :size :accessor size)
   (birth-year :initarg :birth-year :initform 1970 :reader birth-year)))

(defun make-human (name size &rest keys &key birth-year)
  (declare (ignore birth-year))
  (let ((human (apply #'make-instance 'human :name name :size size keys)))
    ;; #### NOTE: as in Java, this count is not accurate because of automatic
    ;; memory management (GC). There's not even a finalize method as in JAVA
    ;; (which is shaky anyway). One would need to implement a protocol such as
    ;; KILL-HUMAN to decrease the count explicitly.
    ;; #### FIXME: obviously, this count is wrong if someone uses
    ;; MAKE-INSTANCE directly.
    (incf (slot-value human 'population))
    human))

;; #### NOTE: CENSUS is a regular function.
;; #### FIXME: we need to pass a HUMAN argument to CENSUS because there's no
;; way to access a shared slot without an instance in CLOS.
(defun census (human)
  (let ((census (slot-value human 'population)))
    (if (zerop census)
	(format t "No human currently alive.~%")
	(format t "~S human~:P currently alive.~%" census)))
  (values))

;; #### NOTE: AGE is a regular function.
(defun age (human)
  (multiple-value-bind (sc mn hr dy mt yr) (get-decoded-time)
    (declare (ignore sc mn hr dy mt))
    (- yr (birth-year human))))

;; In MOB1, we started with HELLO being a regular method. Now we know better,
;; so let's make this directly a generic function.
(defgeneric hello (object))

(defmethod hello ((human human))
  (format t "Hello! I'm ~A, ~Am, ~Ayo.~%"
    (name human)
    (size human)
    (age human))
  (values))


(defclass employee (human)
  ((company :initarg :company :reader company)
   (salary :initarg :salary :accessor salary)
   (hiring-year :initarg :hiring-year)))

(defun make-employee (name size company salary hiring-year
		      &rest keys &key birth-year)
  (declare (ignore birth-year))
  (let ((employee (apply #'make-instance 'employee
			 :name name :size size
			 :company company :salary salary
			 :hiring-year hiring-year
			 keys)))
    ;; #### FIXME: code duplication!
    (incf (slot-value employee 'population))
    employee))

(defun hiring-age (employee)
  (- (slot-value employee 'hiring-year) (birth-year employee)))

(defmethod hello ((employee employee))
  (call-next-method)
  (format t "Working at ~A for ~A euros, started at the age of ~A.~%"
    (company employee)
    (salary employee)
    (hiring-age employee))
  (values))

(defun test ()
  (let (h1 h2 e)
    (setq h1 (make-human "Alain Terieur" 1.80))
    (census h1)
    (hello h1)

    (setq e (make-employee "Alex Terieur" 1.78 "EPITA" 2400 2001))
    (census e)
    (hello e)

    (setq h2 (make-human "Vladimir Guez" 1.85 :birth-year 1980))
    (census h2)
    (format t "~A wants a new name...~%" (name h2))
    (rename "Thierry Chmonfiss" h2)
    (hello h2)))
