;; We're in the CL-USER package by default.
(defclass transport ()
  ((weight :initarg :weight :accessor weight)))


;; SPEED is a CL symbol, so let's not confuse everyone and use VELOCITY
;; instead...

(defpackage boat
  (:use :cl)
  (:export :boat :velocity))

(in-package :boat)

(defclass boat (cl-user::transport)
  ;; This defines both a slot and a generic function named VELOCITY in the
  ;; BOAT package.
  ((velocity :initarg :velocity :accessor velocity)))


(in-package :cl-user)


(defpackage plane
  (:use :cl)
  (:export :plane :velocity))

(in-package :plane)

(defclass plane (cl-user::transport)
  ;; This defines both a slot and a generic function named VELOCITY in the
  ;; PLANE package.
  ((velocity :initarg :velocity :accessor velocity)))


(in-package :cl-user)

(defclass seaplane (boat:boat plane:plane)
  ;; These new initargs and accessors make things clearer when instantiating a
  ;; seaplane. Note that if we use a :velocity initarg, it will serve to
  ;; instantiate both VELOCITY slots.
  ((boat:velocity  :initarg :boat-velocity  :accessor boat-velocity)
   (plane:velocity :initarg :plane-velocity :accessor plane-velocity)))

;; We can also define our own generic wrapper around the package-specific
;; accessors:
(defgeneric velocity (object)
  (:method ((boat boat:boat))    (boat:velocity  boat))
  (:method ((plane plane:plane)) (plane:velocity plane)))


(defun test ()
  (let ((boat  (make-instance 'boat:boat   :weight 1 :velocity 60))
	(plane (make-instance 'plane:plane :weight 2 :velocity 120))
	(seaplane (make-instance 'seaplane :weight 3
					   :boat-velocity 60
					   :plane-velocity 120)))
    (format t "The boat weights ~At and sails at ~Akm/h.~%"
      (weight boat) (velocity boat))
    (format t "The plane weights ~At and flies at ~Akm/h.~%"
      (weight plane) (velocity plane))
    (format t "The seaplane weights ~At, sails at ~Akm/h and flies at ~Akm/h.~%"
      (weight seaplane) (boat-velocity seaplane) (plane-velocity seaplane))))
